
import pytest
from snakestream.collector import to_list
from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_sequential_simple(int_2_letter) -> None:
    # when
    it = await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .sequential() \
        .map(lambda x: int_2_letter[x]) \
        .distinct() \
        .collect(to_list)
    # then
    assert 4 == len(it)
    assert 'a' in it
    assert 'b' in it
    assert 'c' in it
    assert 'd' in it


@pytest.mark.asyncio
async def test_sequential_switch_to_parallel(int_2_letter) -> None:
    # when
    it = await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .parallel() \
        .map(lambda x: int_2_letter[x]) \
        .sequential() \
        .distinct() \
        .collect(to_list)
    # then
    assert 4 == len(it)
    assert 'a' in it
    assert 'b' in it
    assert 'c' in it
    assert 'd' in it



import time
import pytest
import asyncio

from snakestream import Stream
from snakestream.collector import to_list


@pytest.mark.asyncio
async def test_parallel_simple(int_2_letter) -> None:
    # when
    it = await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .parallel() \
        .map(lambda x: int_2_letter[x]) \
        .distinct() \
        .collect(to_list)
    # then
    assert 4 == len(it)
    assert 'a' in it
    assert 'b' in it
    assert 'c' in it
    assert 'd' in it


@pytest.mark.asyncio
async def test_parallel_is_faster_than_sequential() -> None:

    async def sleep(n):
        await asyncio.sleep(0.1)

    # when
    start_parallel = time.time()
    await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .parallel() \
        .map(sleep) \
        .distinct() \
        .collect(to_list)
    end_parallel = time.time()
    time_parallel = end_parallel - start_parallel

    start_sequential = time.time()
    await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .map(sleep) \
        .distinct() \
        .collect(to_list)
    end_sequential = time.time()
    time_sequential = end_sequential - start_sequential

    # then
    # usually something like 0,2 compared to 0.8
    assert (time_parallel) < (time_sequential)


@pytest.mark.asyncio
async def test_sequential_switch_to_sequential(int_2_letter) -> None:
    # when
    it = await Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .sequential() \
        .map(lambda x: int_2_letter[x]) \
        .parallel() \
        .distinct() \
        .collect(to_list)
    # then
    assert 4 == len(it)
    assert 'a' in it
    assert 'b' in it
    assert 'c' in it
    assert 'd' in it



"""
    Dummy conftest.py for snakestream.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""
import asyncio
import sys

import pytest

sys.path.append("src")


class MyObject:
    def __init__(self, identifier, name):
        self.id = identifier
        self.name = name

    def __eq__(self, other):
        if isinstance(other, MyObject):
            return self.id == other.id and self.name == other.name
        return False

    def __hash__(self):
        return hash((self.id, self.name))


@pytest.fixture
def int_2_letter():
    return {
        1: 'a',
        2: 'b',
        3: 'c',
        4: 'd',
        5: 'e',
    }


@pytest.fixture
def letter_2_int(int_2_letter):
    return {v: k for k, v in int_2_letter.items()}


@pytest.fixture(scope='function')
async def async_int_to_letter(int_2_letter):
    async def inner(x: int) -> str:
        await asyncio.sleep(0.01)
        return int_2_letter[x]
    return inner



import pytest
import asyncio

from snakestream import Stream
from snakestream.collector import to_generator
from snakestream.exception import StreamBuildException


async def async_flat_map(x: int) -> int:
    await asyncio.sleep(0.01)
    return x


@pytest.mark.asyncio
async def test_flat_map() -> None:
    # when
    it = Stream.of([[1, 2], [3, 4]]) \
        .flat_map(lambda x: Stream.of(x)) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_flat_map_mixed_list() -> None:
    it = Stream.of([[1, 2], [3, 4], 5, [6, 7], 8]) \
        .flat_map(lambda x: Stream.of(x)) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    assert await it.__anext__() == 5
    assert await it.__anext__() == 6
    assert await it.__anext__() == 7
    assert await it.__anext__() == 8
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_flat_map_async_function() -> None:
    # when
    try:
        Stream.of([[1, 2], [3, 4], 5]) \
            .flat_map(async_flat_map) \
            .collect(to_generator)
    except StreamBuildException:
        pass
    else:
        assert False



from __future__ import annotations

from functools import cmp_to_key
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Callable, Optional, List, \
    Union, AsyncGenerator, Any

from snakestream.base_stream import BaseStream
from snakestream.collector import to_generator
from snakestream.exception import StreamBuildException
from snakestream.sort import merge_sort
from snakestream.type import R, T, Accumulator, CloseHandler, Comparator, Consumer, \
    FlatMapper, Mapper, Predicate


if TYPE_CHECKING:
    from stream_builder import StreamBuilder


PROCESSES: int = 4


async def _concat(a: 'Stream', b: 'Stream') -> AsyncGenerator:
    async for i in a._compose():
        yield i
    async for j in b._compose():
        yield j


class Stream(BaseStream):
    def __init__(self, source: Any, close_handlers: Optional[List[CloseHandler]] = None) -> None:
        super().__init__(source)
        self._close_handlers = close_handlers or []

    @staticmethod
    def of(source: Any) -> 'Stream':
        return Stream(source)

    @staticmethod
    def empty() -> 'Stream':
        return Stream([])

    @staticmethod
    async def concat(a: 'Stream', b: 'Stream') -> 'Stream':
        new_stream = _concat(a, b)
        return Stream(new_stream)

    @staticmethod
    def builder() -> 'StreamBuilder':
        from snakestream.stream_builder import StreamBuilder
        return StreamBuilder()

    # Intermediaries
    def filter(self, predicate: Predicate) -> 'Stream':
        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            async for i in iterable:
                if iscoroutinefunction(predicate):
                    keep = await predicate(i)
                else:
                    keep = predicate(i)
                if keep:
                    yield i

        self._chain.append(fn)
        return self

    def map(self, mapper: Mapper) -> 'Stream':
        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            async for i in iterable:
                if iscoroutinefunction(mapper):
                    yield await mapper(i)
                else:
                    yield mapper(i)

        self._chain.append(fn)
        return self

    def flat_map(self, flat_mapper: FlatMapper) -> 'Stream':
        if iscoroutinefunction(flat_mapper):
            raise StreamBuildException("flat_map() does not support coroutines")

        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            async for i in iterable:
                async for j in flat_mapper(i).collect(to_generator):
                    yield j

        self._chain.append(fn)
        return self

    def sorted(self, comparator: Optional[Comparator] = None, reverse=False) -> 'Stream':
        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            # unfortunately I now don't see other way than to block the entire stream
            # how can I otherwise know what is the first item out?
            cache = []
            async for i in iterable:
                cache.append(i)
            # sort
            if comparator is not None:
                if iscoroutinefunction(comparator):
                    cache = await merge_sort(cache, comparator)
                else:
                    cache.sort(key=cmp_to_key(comparator))
            else:
                cache.sort()
            # unblock the stream
            if reverse:
                for n in reversed(cache):
                    yield n
            else:
                for n in cache:
                    yield n

        self._chain.append(fn)
        return self

    def distinct(self) -> 'Stream':
        seen = set()

        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            async for i in iterable:
                if i in seen:
                    continue
                else:
                    seen.add(i)
                    yield i

        self._chain.append(fn)
        return self

    def peek(self, consumer: Consumer) -> 'Stream':
        async def fn(iterable: AsyncGenerator) -> AsyncGenerator:
            async for i in iterable:
                if iscoroutinefunction(consumer):
                    await consumer(i)
                else:
                    consumer(i)
                yield i

        self._chain.append(fn)
        return self

    # Terminals
    def collect(self, collector: Callable) -> Union[List, AsyncGenerator]:
        return collector(self._compose())

    async def reduce(self, identity: Union[T, R], accumulator: Accumulator) -> Union[T, R]:
        async for n in self._compose():
            if iscoroutinefunction(accumulator):
                identity = await accumulator(identity, n)
            else:
                identity = accumulator(identity, n)
        return identity

    async def for_each(self, consumer: Callable[[T], Any]) -> None:
        async for n in self._compose():
            if iscoroutinefunction(consumer):
                await consumer(n)
            else:
                consumer(n)
        return None

    '''
    async def find_first(self) -> Optional[Any]:
        # until we have ordered parallel stream then we
        # cant do this one
        return await self.find_any()
    '''

    async def find_any(self) -> Optional[Any]:
        async for n in self._compose():
            return n

    async def max(self, comparator: Comparator) -> Optional[T]:
        return await self._min_max(comparator)

    async def min(self, comparator: Comparator) -> Optional[T]:
        if iscoroutinefunction(comparator):
            async def negative_comparator(x, y):
                return not await comparator(x, y)
            return await self._min_max(negative_comparator)
        else:
            def negative_comparator(x, y):
                return not comparator(x, y)
            return await self._min_max(negative_comparator)

    async def _min_max(self, comparator: Comparator) -> Optional[T]:
        found = None
        async for n in self._compose():
            if found is None:
                found = n
                continue

            if iscoroutinefunction(comparator):
                if n and await comparator(n, found):
                    found = n
            else:
                if n and comparator(n, found):
                    found = n
        return found

    async def all_match(self, predicate: Predicate) -> bool:
        async for n in self._compose():
            if iscoroutinefunction(predicate):
                if await predicate(n):
                    continue
                else:
                    return False
            else:
                if predicate(n):
                    continue
                else:
                    return False
        return True

    async def none_match(self, predicate: Predicate) -> bool:
        async for n in self._compose():
            if iscoroutinefunction(predicate):
                if await predicate(n):
                    return False
                else:
                    continue
            else:
                if predicate(n):
                    return False
                else:
                    continue
        return True

    async def any_match(self, predicate: Predicate) -> bool:
        async for n in self._compose():
            if iscoroutinefunction(predicate):
                if await predicate(n):
                    return True
                else:
                    continue
            else:
                if predicate(n):
                    return True
                else:
                    continue
        return False

    async def count(self) -> int:
        c = 0
        async for _ in self._compose():
            c += 1
        return c



import pytest

from snakestream import Stream
from snakestream.collector import to_generator


@pytest.mark.asyncio
async def test_mixed_chain(int_2_letter) -> None:
    # when
    it = Stream.of([1, 2, 3, 4, 5, 6]) \
        .filter(lambda x: 3 < x < 6) \
        .map(lambda x: int_2_letter[x]) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 'd'
    assert await it.__anext__() == 'e'
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False



from typing import TYPE_CHECKING, Any, AsyncGenerator, AsyncIterable, Callable, List, Optional

from snakestream.type import CloseHandler

if TYPE_CHECKING:
    from snakestream.stream import Stream


async def _normalize(source: Any) -> AsyncGenerator:
    if hasattr(source, '__iter__') or hasattr(source, '__next__'):
        for i in source:
            yield i
    else:
        yield source


def _accept(source: Any) -> Optional[AsyncGenerator]:
    if isinstance(source, AsyncGenerator) or isinstance(source, AsyncIterable):
        return source
    return None


class BaseStream():
    def __init__(self, source: Any) -> None:
        self._stream = _accept(source) or _normalize(source)
        self._chain: List[Callable] = []
        self.is_parallel = False
        self._close_handlers = []

    def _sequential(self, intermediaries: List[Callable], iterable: AsyncGenerator) -> AsyncGenerator:
        if len(intermediaries) == 0:
            return iterable
        if len(intermediaries) == 1:
            fn = intermediaries.pop(0)
            return fn(iterable)
        fn = intermediaries.pop(0)
        return self._sequential(intermediaries, fn(iterable))

    def _compose(self) -> AsyncGenerator:
        return self._sequential(self._chain, self._stream)

    def sequential(self) -> 'Stream':
        from .stream import Stream
        new_source = self._compose()
        return Stream(new_source, self._close_handlers)

    def parallel(self) -> 'Stream':
        from .parallel_stream import ParallelStream
        new_source = self._compose()
        return ParallelStream(new_source, self._close_handlers)

    def on_close(self, close_handler: CloseHandler) -> 'Stream':
        self._close_handlers.append(close_handler)
        return self

    def close(self) -> None:
        for close_handler in self._close_handlers:
            close_handler()



async def merge_sort(arr, comparator):
    if len(arr) <= 1:
        return arr

    middle = len(arr) // 2
    left = await merge_sort(arr[:middle], comparator)
    right = await merge_sort(arr[middle:], comparator)

    return await _merge(left, right, comparator)


async def _merge(left, right, comparator):
    result = []
    i = 0
    j = 0
    while i < len(left) and j < len(right):
        if await comparator(left[i], right[j]) <= 0:
            result.append(left[i])
            i += 1
        else:
            result.append(right[j])
            j += 1

    result.extend(left[i:])
    result.extend(right[j:])
    return result



import asyncio

import pytest

from snakestream import Stream


@pytest.mark.asyncio
async def test_empty_stream() -> None:
    it = await Stream.of([]) \
        .none_match(lambda x: x > 5)
    assert it is True


@pytest.mark.asyncio
async def test_simple() -> None:
    it = await Stream.of([1, 2, 3, 2, 3, 1, 2]) \
        .none_match(lambda x: x > 5)
    assert it is True


@pytest.mark.asyncio
async def test_some_matches() -> None:
    it = await Stream.of([1, 2, 3, 2, 3, 1, 2, 5, 6, 7]) \
        .none_match(lambda x: x < 5)
    assert it is False


@pytest.mark.asyncio
async def test_simple_async() -> None:
    async def async_predicate(x: int) -> bool:
        await asyncio.sleep(0.01)
        return x > 5

    it = await Stream.of([1, 2, 3, 2, 3, 1, 2]) \
        .none_match(async_predicate)
    assert it is True


@pytest.mark.asyncio
async def test_all_matches() -> None:
    it = await Stream.of([1, 2, 3, 2, 3, 1, 2]) \
        .none_match(lambda x: x < 5)
    assert it is False



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name

from typing import AsyncGenerator, Any, List


async def to_generator(composition: AsyncGenerator) -> AsyncGenerator[Any, None]:
    async for n in composition:
        yield n


async def to_list(composition: AsyncGenerator) -> List[Any]:
    ret = []
    async for n in composition:
        ret.append(n)
    return ret



import asyncio
from typing import Any, AsyncGenerator, Callable, List, Optional
from snakestream.stream import PROCESSES, Stream
from snakestream.type import CloseHandler


class ParallelStream(Stream):
    def __init__(self, source: Any, close_handlers: Optional[List[CloseHandler]] = None) -> None:
        super().__init__(source)
        self._close_handlers = close_handlers or []

    def _compose(self) -> AsyncGenerator:
        return self._parallel(self._chain, self._stream)

    async def _parallel(
        self,
        intermediaries: List[Callable],
        iterable: AsyncGenerator,
        processes: int = PROCESSES
    ) -> AsyncGenerator:
        async_iterators = [self._sequential(intermediaries[:], iterable) for n in range(processes)]
        tasks = [asyncio.ensure_future(n.__anext__()) for n in async_iterators]

        while any([n is not None for n in tasks]):

            waitlist = filter(lambda n: n is not None, tasks)
            done, _ = await asyncio.wait(waitlist, return_when=asyncio.FIRST_COMPLETED)

            for task in done:
                task_idx = tasks.index(task)
                try:
                    result = tasks[task_idx].result()
                    tasks[task_idx] = asyncio.ensure_future(async_iterators[task_idx].__anext__())
                    yield result
                except StopAsyncIteration:
                    tasks[task_idx] = None



import asyncio

import pytest

from conftest import MyObject
from snakestream import Stream
from snakestream.collector import to_list

obj1 = MyObject(1, "Object1")
obj2 = MyObject(2, "Object2")
obj3 = MyObject(3, "Object3")


input_list = [
    obj1,
    obj2,
    obj3,
    obj2,
    obj3,
]


@pytest.mark.asyncio
async def test_ok() -> None:
    # when
    it = await Stream.of(input_list) \
        .peek(lambda x: x) \
        .collect(to_list)
    # then
    assert it == input_list
    assert it[0] is input_list[0]
    assert it[1] is input_list[1]
    assert it[2] is input_list[2]
    assert it[3] is input_list[3]
    assert it[4] is input_list[4]


@pytest.mark.asyncio
async def test_ok_async_function() -> None:
    names = []

    async def some_func(x: MyObject) -> None:
        await asyncio.sleep(0.01)
        names.append(x.name)
        return

    # when
    it = await Stream.of(input_list) \
        .peek(some_func) \
        .collect(to_list)
    # then
    assert it == input_list
    assert it[0] is input_list[0]
    assert it[1] is input_list[1]
    assert it[2] is input_list[2]
    assert it[3] is input_list[3]
    assert it[4] is input_list[4]

    assert names == ['Object1', 'Object2', 'Object3', 'Object2', 'Object3',]


@pytest.mark.asyncio
async def test_empty_stream() -> None:
    # when
    it = await Stream.of([]) \
        .peek(lambda x: x) \
        .collect(to_list)
    # then
    assert it == []


@pytest.mark.asyncio
async def test_multiple_calls() -> None:
    # when
    it = await Stream.of(input_list) \
        .peek(lambda x: x) \
        .peek(lambda x: x) \
        .collect(to_list)
    # then
    assert it == input_list
    assert it[0] is input_list[0]
    assert it[1] is input_list[1]
    assert it[2] is input_list[2]
    assert it[3] is input_list[3]
    assert it[4] is input_list[4]


@pytest.mark.asyncio
async def test_mutate_internal_state() -> None:
    def lower_name(x: MyObject) -> None:
        x.name = x.name.lower()

    # when
    it = await Stream.of(input_list) \
        .peek(lower_name) \
        .collect(to_list)
    # then
    assert it == input_list
    assert it[0] is input_list[0]
    assert it[1] is input_list[1]
    assert it[2] is input_list[2]
    assert it[3] is input_list[3]
    assert it[4] is input_list[4]

    assert it[0].name == 'object1'
    assert it[1].name == 'object2'
    assert it[2].name == 'object3'
    assert it[3].name == 'object2'
    assert it[4].name == 'object3'



from typing import TYPE_CHECKING, AsyncGenerator, \
    AsyncIterable, Awaitable, Callable, Generator, Iterable, Optional, TypeVar, Union

if TYPE_CHECKING:
    from snakestream.stream import Stream

T = TypeVar('T')
R = TypeVar('R')

Predicate = Callable[[T], Union[bool, Awaitable[bool]]]

# Intermediaries
Filterer = Callable[[T], T]
Mapper = Callable[[T], Optional[R]]
FlatMapper = Callable[[Union[Iterable, AsyncIterable, Generator, AsyncGenerator]], 'Stream']
Comparator = Callable[[T, T], Union[bool, Awaitable[bool]]]
Consumer = Callable[[T], T]
CloseHandler = Callable[[], None]

# Terminals
Accumulator = Callable[[T, Union[T, R]], Union[T, R]]
